import json
from pathlib import Path
from typing import Generator

from .constants import PatchPlanFilename
from .models import EdgeToAdd, EdgeToDelete, EdgeToUpdate, PatchPlan, VertexToAdd, VertexToDelete, VertexToUpdate


class PatchPlanReader:
    def read(self, patch_plan_directory: Path) -> PatchPlan:
        vertices_to_add = self._read_vertices_to_add(patch_plan_directory=patch_plan_directory)
        vertices_to_delete = self._read_vertices_to_delete(patch_plan_directory=patch_plan_directory)
        vertices_to_update = self._read_vertices_to_update(patch_plan_directory=patch_plan_directory)
        edges_to_add = self._read_edges_to_add(patch_plan_directory=patch_plan_directory)
        edges_to_delete = self._read_edges_to_delete(patch_plan_directory=patch_plan_directory)
        edges_to_update = self._read_edges_to_update(patch_plan_directory=patch_plan_directory)
        added_node_db_id_map = self._read_added_node_db_id_map(patch_plan_directory=patch_plan_directory)
        deleted_db_ids = self._read_deleted_db_ids(patch_plan_directory=patch_plan_directory)
        reverted_deleted_db_id_map = self._read_reverted_deleted_db_id_map(patch_plan_directory=patch_plan_directory)

        return PatchPlan(
            name="none",
            vertices_to_add=vertices_to_add,
            vertices_to_delete=vertices_to_delete,
            vertices_to_update=vertices_to_update,
            edges_to_add=edges_to_add,
            edges_to_delete=edges_to_delete,
            edges_to_update=edges_to_update,
            added_element_db_id_map=added_node_db_id_map or {},
            deleted_db_ids=deleted_db_ids or set(),
            reverted_deleted_db_id_map=reverted_deleted_db_id_map or {},
        )

    def _read_file_lines(self, patch_file: Path) -> Generator[str | None, None, None]:
        if not patch_file.exists():
            return
        with patch_file.open() as f:
            yield from f

    def _read_vertices_to_add(self, patch_plan_directory: Path) -> list[VertexToAdd]:
        file = patch_plan_directory / Path(PatchPlanFilename.VERTICES_TO_ADD.value)
        vertices_to_add: list[VertexToAdd] = []
        for raw_line in self._read_file_lines(patch_file=file):
            if raw_line:
                vertices_to_add.append(VertexToAdd(**json.loads(raw_line)))
        return vertices_to_add

    def _read_vertices_to_update(self, patch_plan_directory: Path) -> list[VertexToUpdate]:
        file = patch_plan_directory / Path(PatchPlanFilename.VERTICES_TO_UPDATE.value)
        vertices_to_update: list[VertexToUpdate] = []
        for raw_line in self._read_file_lines(patch_file=file):
            if raw_line:
                vertices_to_update.append(VertexToUpdate(**json.loads(raw_line)))
        return vertices_to_update

    def _read_vertices_to_delete(self, patch_plan_directory: Path) -> list[VertexToDelete]:
        file = patch_plan_directory / Path(PatchPlanFilename.VERTICES_TO_DELETE.value)
        vertices_to_delete: list[VertexToDelete] = []
        for raw_line in self._read_file_lines(patch_file=file):
            if raw_line:
                vertices_to_delete.append(VertexToDelete(**json.loads(raw_line)))
        return vertices_to_delete

    def _read_edges_to_add(self, patch_plan_directory: Path) -> list[EdgeToAdd]:
        file = patch_plan_directory / Path(PatchPlanFilename.EDGES_TO_ADD.value)
        edges_to_add: list[EdgeToAdd] = []
        for raw_line in self._read_file_lines(patch_file=file):
            if raw_line:
                edges_to_add.append(EdgeToAdd(**json.loads(raw_line)))
        return edges_to_add

    def _read_edges_to_delete(self, patch_plan_directory: Path) -> list[EdgeToDelete]:
        file = patch_plan_directory / Path(PatchPlanFilename.EDGES_TO_DELETE.value)
        edges_to_delete: list[EdgeToDelete] = []
        for raw_line in self._read_file_lines(patch_file=file):
            if raw_line:
                edges_to_delete.append(EdgeToDelete(**json.loads(raw_line)))
        return edges_to_delete

    def _read_edges_to_update(self, patch_plan_directory: Path) -> list[EdgeToUpdate]:
        file = patch_plan_directory / Path(PatchPlanFilename.EDGES_TO_UPDATE.value)
        edges_to_update: list[EdgeToUpdate] = []
        for raw_line in self._read_file_lines(patch_file=file):
            if raw_line:
                edges_to_update.append(EdgeToUpdate(**json.loads(raw_line)))
        return edges_to_update

    def _read_added_node_db_id_map(self, patch_plan_directory: Path) -> dict[str, str] | None:
        file = patch_plan_directory / Path(PatchPlanFilename.ADDED_DB_IDS.value)
        if not file.exists():
            return None
        added_db_id_json = file.read_text()
        return json.loads(added_db_id_json)

    def _read_deleted_db_ids(self, patch_plan_directory: Path) -> set[str] | None:
        file = patch_plan_directory / Path(PatchPlanFilename.DELETED_DB_IDS.value)
        if not file.exists():
            return None
        deleted_db_ids_json = file.read_text()
        return set(json.loads(deleted_db_ids_json))

    def _read_reverted_deleted_db_id_map(self, patch_plan_directory: Path) -> dict[str, str] | None:
        file = patch_plan_directory / Path(PatchPlanFilename.REVERTED_DELETED_DB_IDS.value)
        if not file.exists():
            return None
        reverted_deleted_db_id_json = file.read_text()
        return json.loads(reverted_deleted_db_id_json)



from dataclasses import dataclass, field

from infrahub.core.constants import DiffAction, RelationshipCardinality

from .model.path import EnrichedDiffNode, EnrichedDiffRelationship, EnrichedDiffRoot, NodeIdentifier


@dataclass
class ParentNodeAddRequest:
    node_identifier: NodeIdentifier
    parent_identifier: NodeIdentifier
    parent_label: str
    parent_rel_name: str
    parent_rel_identifier: str
    parent_rel_cardinality: RelationshipCardinality
    parent_rel_label: str = field(default="")


class DiffParentNodeAdder:
    def __init__(self) -> None:
        self._diff_root: EnrichedDiffRoot | None = None
        self._node_map: dict[NodeIdentifier, EnrichedDiffNode] = {}

    def initialize(self, enriched_diff_root: EnrichedDiffRoot) -> None:
        self._diff_root = enriched_diff_root
        self._node_map = enriched_diff_root.get_node_map()

    def get_root(self) -> EnrichedDiffRoot:
        if not self._diff_root:
            raise RuntimeError("Must call initialize before using")
        return self._diff_root

    def get_node(self, identifier: NodeIdentifier) -> EnrichedDiffNode:
        return self._node_map[identifier]

    def has_node(self, identifier: NodeIdentifier) -> bool:
        return identifier in self._node_map

    def add_node(self, node: EnrichedDiffNode) -> None:
        if node.identifier in self._node_map:
            return
        self._node_map[node.identifier] = node
        self.get_root().nodes.add(node)

    def add_parent(self, parent_request: ParentNodeAddRequest) -> EnrichedDiffNode:
        if not self._diff_root:
            raise RuntimeError("Must call initialize before using")
        node = self.get_node(identifier=parent_request.node_identifier)
        if not self.has_node(identifier=parent_request.parent_identifier):
            parent = EnrichedDiffNode(
                identifier=parent_request.parent_identifier,
                label=parent_request.parent_label,
                action=DiffAction.UNCHANGED,
                changed_at=None,
            )
            self.add_node(parent)
        else:
            parent = self.get_node(identifier=parent_request.parent_identifier)

        try:
            rel = node.get_relationship(name=parent_request.parent_rel_name)
            rel.nodes.add(parent)
        except ValueError:
            node.relationships.add(
                EnrichedDiffRelationship(
                    name=parent_request.parent_rel_name,
                    identifier=parent_request.parent_rel_identifier,
                    label=parent_request.parent_rel_label,
                    cardinality=parent_request.parent_rel_cardinality,
                    changed_at=None,
                    action=DiffAction.UNCHANGED,
                    nodes={parent},
                )
            )

        return parent



from enum import Enum


class PatchPlanFilename(str, Enum):
    VERTICES_TO_ADD = "vertices_to_add.json"
    VERTICES_TO_UPDATE = "vertices_to_update.json"
    VERTICES_TO_DELETE = "vertices_to_delete.json"
    EDGES_TO_ADD = "edges_to_add.json"
    EDGES_TO_UPDATE = "edges_to_update.json"
    EDGES_TO_DELETE = "edges_to_delete.json"
    ADDED_DB_IDS = "added_db_ids.json"
    DELETED_DB_IDS = "deleted_db_ids.json"
    REVERTED_DELETED_DB_IDS = "reverted_deleted_db_ids.json"



import asyncio

from prefect.server.database import provide_database_interface
from prefect.server.models.block_registration import run_block_auto_registration


async def init_prefect() -> None:
    db = provide_database_interface()

    await db.create_db()
    session = await db.session()

    async with session:
        await run_block_auto_registration(session=session)


if __name__ == "__main__":
    asyncio.run(init_prefect())



from graphene import InputObjectType

from infrahub.core.branch import Branch
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.schema import MainSchemaTypes
from infrahub.database import InfrahubDatabase

from .interface import MutationNodeGetterInterface


class MutationNodeGetterById(MutationNodeGetterInterface):
    def __init__(self, db: InfrahubDatabase, node_manager: NodeManager) -> None:
        self.db = db
        self.node_manager = node_manager

    async def get_node(
        self,
        node_schema: MainSchemaTypes,
        data: InputObjectType,
        branch: Branch,
    ) -> Node | None:
        node = None
        if "id" not in data:
            return node
        return await self.node_manager.get_one(id=str(data["id"]), db=self.db, branch=branch, kind=node_schema.kind)



import asyncio
import logging
import os
import signal
from typing import TYPE_CHECKING, Any

import typer
from infrahub_sdk import Config, InfrahubClient
from infrahub_sdk.async_typer import AsyncTyper
from infrahub_sdk.exceptions import Error as SdkError
from prometheus_client import start_http_server
from rich.logging import RichHandler

from infrahub import __version__, config
from infrahub.components import ComponentType
from infrahub.core.initialization import initialization
from infrahub.dependencies.registry import build_component_registry
from infrahub.git import initialize_repositories_directory
from infrahub.lock import initialize_lock
from infrahub.log import get_logger
from infrahub.services import InfrahubServices
from infrahub.services.adapters.cache import InfrahubCache
from infrahub.services.adapters.message_bus import InfrahubMessageBus
from infrahub.services.adapters.workflow.local import WorkflowLocalExecution
from infrahub.services.adapters.workflow.worker import WorkflowWorkerExecution
from infrahub.trace import configure_trace

if TYPE_CHECKING:
    from infrahub.cli.context import CliContext

app = AsyncTyper()
log = get_logger()

shutdown_event = asyncio.Event()


def signal_handler(*args: Any, **kwargs: Any) -> None:  # noqa: ARG001
    shutdown_event.set()


signal.signal(signal.SIGINT, signal_handler)


@app.callback()
def callback() -> None:
    """
    Control the Git Agent.
    """


async def initialize_git_agent(service: InfrahubServices) -> None:
    service.log.info("Initializing Git Agent ...")
    initialize_repositories_directory()


@app.command()
async def start(
    ctx: typer.Context,
    debug: bool = typer.Option(False, help="Enable advanced logging and troubleshooting"),
    config_file: str = typer.Option(
        "infrahub.toml", envvar="INFRAHUB_CONFIG", help="Location of the configuration file to use for Infrahub"
    ),
    port: int = typer.Argument(8000, envvar="INFRAHUB_METRICS_PORT", help="Port used to expose a metrics endpoint"),
) -> None:
    """Start Infrahub Git Agent."""
    logging.getLogger("httpx").setLevel(logging.ERROR)
    logging.getLogger("httpcore").setLevel(logging.ERROR)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("aio_pika").setLevel(logging.ERROR)
    logging.getLogger("aiormq").setLevel(logging.ERROR)
    logging.getLogger("git").setLevel(logging.ERROR)
    logging.getLogger("aiosqlite").setLevel(logging.ERROR)

    log.debug(f"Config file : {config_file}")
    # Prevent git from interactively prompting the user for passwords if the credentials provided
    # by the credential helper is failing.
    os.environ["GIT_TERMINAL_PROMPT"] = "0"

    context: CliContext = ctx.obj

    config.load_and_exit(config_file_name=config_file)

    log_level = "DEBUG" if debug else "INFO"

    FORMAT = "%(name)s | %(message)s" if debug else "%(message)s"
    logging.basicConfig(level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])

    # Start the metrics endpoint
    start_http_server(port)

    # initialize the Infrahub Client and query the list of branches to validate that the API is reacheable and the auth is working
    log.debug(f"Using Infrahub API at {config.SETTINGS.main.internal_address}")
    client = InfrahubClient(
        config=Config(address=config.SETTINGS.main.internal_address, retry_on_failure=True, log=log)
    )
    try:
        await client.branch.all()
    except SdkError as exc:
        log.error(f"Error in communication with Infrahub: {exc.message}")
        raise typer.Exit(1) from None

    # Initialize trace
    if config.SETTINGS.trace.enable:
        configure_trace(
            service="infrahub-git-agent",
            version=__version__,
            exporter_type=config.SETTINGS.trace.exporter_type,
            exporter_endpoint=config.SETTINGS.trace.exporter_endpoint,
            exporter_protocol=config.SETTINGS.trace.exporter_protocol,
        )

    database = await context.init_db(retry=1)

    workflow = config.OVERRIDE.workflow or (
        WorkflowWorkerExecution()
        if config.SETTINGS.workflow.driver == config.WorkflowDriver.WORKER
        else WorkflowLocalExecution()
    )

    component_type = ComponentType.GIT_AGENT
    message_bus = config.OVERRIDE.message_bus or (
        await InfrahubMessageBus.new_from_driver(component_type=component_type, driver=config.SETTINGS.broker.driver)
    )
    cache = config.OVERRIDE.cache or (await InfrahubCache.new_from_driver(driver=config.SETTINGS.cache.driver))

    service = await InfrahubServices.new(
        cache=cache,
        client=client,
        database=database,
        workflow=workflow,
        message_bus=message_bus,
        component_type=component_type,
    )

    # Initialize the lock
    initialize_lock(service=service)

    async with service.database.start_session() as db:
        await initialization(db=db)

    await service.component.refresh_schema_hash()

    await initialize_git_agent(service=service)

    build_component_registry()

    log.info("Initialized Git Agent ...")

    await shutdown_event.wait()

    log.info("Shutdown of Git agent requested")

    await service.shutdown()
    log.info("All services stopped")



from pydantic import BaseModel

from .constants import InfrahubType


class TelemetryWorkerData(BaseModel):
    total: int
    active: int


class TelemetryBranchData(BaseModel):
    total: int


class TelemetrySchemaData(BaseModel):
    node_count: int
    generic_count: int
    last_update: str


class TelemetryDatabaseServerData(BaseModel):
    name: str
    version: str


class TelemetryDatabaseSystemInfoData(BaseModel):
    memory_total: int
    memory_available: int
    processor_available: int


class TelemetryDatabaseData(BaseModel):
    database_type: str
    relationship_count: dict[str, int]
    node_count: dict[str, int]
    servers: list[TelemetryDatabaseServerData]
    system_info: TelemetryDatabaseSystemInfoData | None


class TelemetryWorkPoolData(BaseModel):
    name: str
    type: str
    total_workers: int
    active_workers: int


class TelemetryPrefectData(BaseModel):
    events: dict[str, int]
    automations: dict[str, int]
    work_pools: list[TelemetryWorkPoolData]


class TelemetryData(BaseModel):
    deployment_id: str | None
    execution_time: float | None
    infrahub_version: str
    infrahub_type: InfrahubType
    python_version: str
    platform: str
    workers: TelemetryWorkerData
    branches: TelemetryBranchData
    features: dict[str, int]
    schema_info: TelemetrySchemaData
    database: TelemetryDatabaseData
    prefect: TelemetryPrefectData



from infrahub.message_bus import InfrahubMessage, InfrahubResponse

from .git_file_get import GitFileGet, GitFileGetResponse
from .git_repository_connectivity import GitRepositoryConnectivity
from .refresh_git_fetch import RefreshGitFetch
from .refresh_registry_branches import RefreshRegistryBranches
from .refresh_registry_rebasedbranch import RefreshRegistryRebasedBranch
from .send_echo_request import SendEchoRequest, SendEchoRequestResponse

MESSAGE_MAP: dict[str, type[InfrahubMessage]] = {
    "git.file.get": GitFileGet,
    "git.repository.connectivity": GitRepositoryConnectivity,
    "refresh.git.fetch": RefreshGitFetch,
    "refresh.registry.branches": RefreshRegistryBranches,
    "refresh.registry.rebased_branch": RefreshRegistryRebasedBranch,
    "send.echo.request": SendEchoRequest,
}

RESPONSE_MAP: dict[str, type[InfrahubResponse]] = {
    "git.file.get": GitFileGetResponse,
    "send.echo.request": SendEchoRequestResponse,
}

PRIORITY_MAP = {
    "check.artifact.create": 2,
    "check.repository.check_definition": 2,
    "check.repository.merge_conflicts": 2,
    "send.echo.request": 5,  # Currently only for testing purposes, will be removed once all message bus have been migrated to prefect
    "event.branch.delete": 5,
    "event.branch.merge": 5,
    "event.schema.update": 5,
    "git.diff.names_only": 4,
    "git.file.get": 4,
    "request.artifact.generate": 2,
    "request.git.sync": 4,
    "request.proposed_change.pipeline": 5,
    "transform.jinja.template": 4,
    "transform.python.data": 4,
}


def message_priority(routing_key: str) -> int:
    return PRIORITY_MAP.get(routing_key, 3)


ROUTING_KEY_MAP: dict[type[InfrahubMessage], str] = {
    message: routing_key for routing_key, message in MESSAGE_MAP.items()
}



from typing import Sequence, cast

from infrahub.core.constants import InfrahubKind, ValidatorConclusion, ValidatorState
from infrahub.core.diff.model.diff import ObjectConflict
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.protocols import CoreProposedChange
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import NodeNotFoundError


class ObjectConflictValidatorRecorder:
    def __init__(self, db: InfrahubDatabase, validator_kind: str, validator_label: str, check_schema_kind: str) -> None:
        self.db = db
        self.validator_kind = validator_kind
        self.validator_label = validator_label
        self.check_schema_kind = check_schema_kind

    async def record_conflicts(self, proposed_change_id: str, conflicts: Sequence[ObjectConflict]) -> list[Node]:
        try:
            proposed_change = await NodeManager.get_one_by_id_or_default_filter(
                id=proposed_change_id, kind=InfrahubKind.PROPOSEDCHANGE, db=self.db
            )
        except NodeNotFoundError:
            return []
        proposed_change = cast(CoreProposedChange, proposed_change)
        validator = await self.get_or_create_validator(proposed_change)
        await self.initialize_validator(validator)

        previous_checks = await validator.checks.get_peers(db=self.db)  # type: ignore[attr-defined]
        previous_checks_by_conflict_id = {check.enriched_conflict_id.value: check for check in previous_checks.values()}
        is_success = False

        current_checks: list[Node] = []
        check_ids_to_keep = set()
        if not conflicts:
            is_success = True
            check = None
            for previous_check in previous_checks.values():
                if previous_check.conflicts.value == []:  # type: ignore[attr-defined]
                    check = previous_check
                    check_ids_to_keep.add(check.id)

            if not check:
                check = await Node.init(db=self.db, schema=self.check_schema_kind)
                await check.new(
                    db=self.db,
                    label="Data Conflict",
                    origin="internal",
                    kind="DataIntegrity",
                    validator=validator.id,
                    conclusion="success",
                    severity="info",
                    conflicts=[],
                )
                await check.save(db=self.db)
            current_checks.append(check)

        for conflict in conflicts:
            conflicts_data = [conflict.to_conflict_dict()]
            conflict_obj = None
            if conflict.conflict_id and conflict.conflict_id in previous_checks_by_conflict_id:
                conflict_obj = previous_checks_by_conflict_id[conflict.conflict_id]
                check_ids_to_keep.add(conflict_obj.get_id())
            if not conflict_obj:
                for previous_check in previous_checks.values():
                    if previous_check.conflicts.value == conflicts_data:  # type: ignore[attr-defined]
                        conflict_obj = previous_check
                        check_ids_to_keep.add(conflict_obj.id)

            if not conflict_obj:
                conflict_obj = await Node.init(db=self.db, schema=self.check_schema_kind)

                await conflict_obj.new(
                    db=self.db,
                    enriched_conflict_id=conflict.conflict_id,
                    label=conflict.label,
                    origin="internal",
                    kind="DataIntegrity",
                    validator=validator.id,
                    conclusion="failure",
                    severity="critical",
                    conflicts=conflicts_data,
                )

                await conflict_obj.save(db=self.db)
            current_checks.append(conflict_obj)

        for check_id, previous_check in previous_checks.items():
            if check_id not in check_ids_to_keep:
                await previous_check.delete(db=self.db)

        await self.finalize_validator(validator, is_success)
        return current_checks

    async def get_validator(self, proposed_change: CoreProposedChange) -> Node | None:
        validations = await proposed_change.validations.get_peers(db=self.db, branch_agnostic=True)

        for validation in validations.values():
            if validation.get_kind() == self.validator_kind:
                return validation
        return None

    async def get_or_create_validator(self, proposed_change: CoreProposedChange) -> Node:
        validator_obj = await self.get_validator(proposed_change=proposed_change)
        if validator_obj:
            return validator_obj
        validator_obj = await Node.init(db=self.db, schema=self.validator_kind)
        await validator_obj.new(
            db=self.db,
            label=self.validator_label,
            state=ValidatorState.QUEUED.value,
            conclusion=ValidatorConclusion.UNKNOWN.value,
            proposed_change=proposed_change.id,
        )
        await validator_obj.save(db=self.db)
        return validator_obj

    async def initialize_validator(self, validator: Node) -> None:
        validator.state.value = ValidatorState.IN_PROGRESS.value  # type: ignore[attr-defined]
        validator.conclusion.value = ValidatorConclusion.UNKNOWN.value  # type: ignore[attr-defined]
        validator.started_at.value = Timestamp().to_string()  # type: ignore[attr-defined]
        validator.completed_at.value = ""  # type: ignore[attr-defined]
        await validator.save(db=self.db)

    async def finalize_validator(self, validator: Node, is_success: bool) -> None:
        validator.state.value = ValidatorState.COMPLETED.value  # type: ignore[attr-defined]
        validator.conclusion.value = (  # type: ignore[attr-defined]
            ValidatorConclusion.SUCCESS.value if is_success else ValidatorConclusion.FAILURE.value
        )
        validator.completed_at.value = Timestamp().to_string()  # type: ignore[attr-defined]
        await validator.save(db=self.db)



from pathlib import Path

from infrahub import config
from infrahub.log import get_logger

log = get_logger()


def get_repositories_directory() -> Path:
    """Return the absolute path to the main directory used for the repositories."""
    return Path(config.SETTINGS.git.repositories_directory).resolve()


def initialize_repositories_directory() -> bool:
    """Check if the main repositories_directory already exist, if not create it.

    Return
        True if the directory has been created,
        False if the directory was already present.
    """
    repos_dir = get_repositories_directory()
    if not repos_dir.exists():
        repos_dir.mkdir(parents=True)

        log.debug(f"Initialized the repositories_directory at {repos_dir}")
        return True

    log.debug(f"repositories_directory already present at {repos_dir}")
    return False



from __future__ import annotations

from collections.abc import Iterable
from typing import Any

import pytest
import yaml
from pytest import Item

from .exceptions import InvalidResourceConfigError
from .items import (
    InfrahubCheckIntegrationItem,
    InfrahubCheckSmokeItem,
    InfrahubCheckUnitProcessItem,
    InfrahubGraphQLQueryIntegrationItem,
    InfrahubGraphQLQuerySmokeItem,
    InfrahubItem,
    InfrahubJinja2TransformIntegrationItem,
    InfrahubJinja2TransformSmokeItem,
    InfrahubJinja2TransformUnitRenderItem,
    InfrahubPythonTransformIntegrationItem,
    InfrahubPythonTransformSmokeItem,
    InfrahubPythonTransformUnitProcessItem,
)
from .models import InfrahubTestFileV1, InfrahubTestGroup

MARKER_MAPPING = {
    "Check": pytest.mark.infrahub_check,
    "GraphQLQuery": pytest.mark.infrahub_graphql_query,
    "Jinja2Transform": pytest.mark.infrahub_jinja2_transform,
    "PythonTransform": pytest.mark.infrahub_python_transform,
}
CONFIG_MAPPING = {
    "Check": "get_check_definition",
    "GraphQLQuery": None,
    "Jinja2Transform": "get_jinja2_transform",
    "PythonTransform": "get_python_transform",
}

ITEMS_MAPPING = {
    "check-smoke": InfrahubCheckSmokeItem,
    "check-unit-process": InfrahubCheckUnitProcessItem,
    "check-integration": InfrahubCheckIntegrationItem,
    "graphql-query-smoke": InfrahubGraphQLQuerySmokeItem,
    "graphql-query-integration": InfrahubGraphQLQueryIntegrationItem,
    "jinja2-transform-smoke": InfrahubJinja2TransformSmokeItem,
    "jinja2-transform-unit-render": InfrahubJinja2TransformUnitRenderItem,
    "jinja2-transform-integration": InfrahubJinja2TransformIntegrationItem,
    "python-transform-smoke": InfrahubPythonTransformSmokeItem,
    "python-transform-unit-process": InfrahubPythonTransformUnitProcessItem,
    "python-transform-integration": InfrahubPythonTransformIntegrationItem,
}


class InfrahubYamlFile(pytest.File):
    def get_resource_config(self, group: InfrahubTestGroup) -> Any | None:
        """Retrieve the resource configuration to apply to all tests in a group."""
        resource_config_function = CONFIG_MAPPING.get(group.resource)

        resource_config = None
        if resource_config_function is not None:
            func = getattr(self.session.infrahub_repo_config, resource_config_function)  # type:ignore[attr-defined]
            try:
                resource_config = func(group.resource_name)
            except KeyError:
                # Ignore error and just return None
                pass

        return resource_config

    def collect_group(self, group: InfrahubTestGroup) -> Iterable[Item]:
        """Collect all items for a group."""
        marker = MARKER_MAPPING[group.resource]
        resource_config = self.get_resource_config(group)

        for test in group.tests:
            item_class: type[pytest.Item] = ITEMS_MAPPING[test.spec.kind]  # type: ignore[assignment]
            item: InfrahubItem = item_class.from_parent(
                name=f"{marker.markname}__{group.resource_name}__{test.name}",
                parent=self,
                resource_name=group.resource_name,
                resource_config=resource_config,
                test=test,
            )  # type: ignore[assignment]

            # If item does not pass validation, mark it to be skipped
            try:
                item.validate_resource_config()
            except InvalidResourceConfigError as exc:
                item.add_marker(pytest.mark.skip(reason=str(exc)))

            item.add_marker(pytest.mark.infrahub)
            item.add_marker(marker)
            if "smoke" in test.spec.kind:
                item.add_marker(pytest.mark.infrahub_smoke)
            if "unit" in test.spec.kind:
                item.add_marker(pytest.mark.infrahub_unit)
            if "integration" in test.spec.kind:
                item.add_marker(pytest.mark.infrahub_integration)

            yield item

    def collect(self) -> Iterable[Item]:
        raw = yaml.safe_load(self.path.open(encoding="utf-8"))

        if "infrahub_tests" not in raw:
            return

        content = InfrahubTestFileV1(**raw)

        for test_group in content.infrahub_tests:
            yield from self.collect_group(test_group)



from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class RefreshRegistryRebasedBranch(InfrahubMessage):
    """Sent to refresh a rebased branch within the local registry."""

    branch: str = Field(..., description="The branch that was rebased")
